---Specialization for automatically load objects onto a vehicle




PalletAutoLoader = {}







---
function PalletAutoLoader.prerequisitesPresent(specializations)
    return true
end


---
function PalletAutoLoader.initSpecialization()
    local schema = Vehicle.xmlSchema
    schema:setXMLSpecializationType("PalletAutoLoader")

    schema:register(XMLValueType.NODE_INDEX, "vehicle.palletAutoLoader.trigger#node", "Trigger node")
    schema:register(XMLValueType.NODE_INDEX, "vehicle.palletAutoLoader.pickupTriggers.pickupTrigger(?)#node", "Pickup trigger node")
    schema:register(XMLValueType.NODE_INDEX, "vehicle.palletAutoLoader.loadPlaces.loadPlace(?)#node", "Load place node")
    schema:register(XMLValueType.VECTOR_TRANS, "vehicle.palletAutoLoader.loadPlaces#checkOffset", "Check offset from load place node")
    schema:register(XMLValueType.VECTOR_TRANS, "vehicle.palletAutoLoader.loadPlaces#checkSize", "Check size around load place node")
    schema:register(XMLValueType.STRING, "vehicle.palletAutoLoader#supportedObject", "Path to xml of supported object")
    schema:register(XMLValueType.INT, "vehicle.palletAutoLoader#fillUnitIndex", "Fill unit index to check fill type")
    schema:register(XMLValueType.INT, "vehicle.palletAutoLoader#maxObjects", "Max. number of objects to load", "Number of load places")
    schema:register(XMLValueType.BOOL, "vehicle.palletAutoLoader#useBales", "Use for bales", false)
    schema:register(XMLValueType.BOOL, "vehicle.palletAutoLoader#useTensionBelts", "Automatically mount tension belts", "False for mobile, otherwise true")

    schema:setXMLSpecializationType()
end


---
function PalletAutoLoader.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "getIsValidObject", PalletAutoLoader.getIsValidObject)
    SpecializationUtil.registerFunction(vehicleType, "getIsAutoLoadingAllowed", PalletAutoLoader.getIsAutoLoadingAllowed)
    SpecializationUtil.registerFunction(vehicleType, "getFirstValidLoadPlace", PalletAutoLoader.getFirstValidLoadPlace)
    SpecializationUtil.registerFunction(vehicleType, "autoLoaderOverlapCallback", PalletAutoLoader.autoLoaderOverlapCallback)
    SpecializationUtil.registerFunction(vehicleType, "autoLoaderTriggerCallback", PalletAutoLoader.autoLoaderTriggerCallback)
    SpecializationUtil.registerFunction(vehicleType, "autoLoaderPickupTriggerCallback", PalletAutoLoader.autoLoaderPickupTriggerCallback)
    SpecializationUtil.registerFunction(vehicleType, "onDeletePalletAutoLoaderObject", PalletAutoLoader.onDeletePalletAutoLoaderObject)
    SpecializationUtil.registerFunction(vehicleType, "loadObject", PalletAutoLoader.loadObject)
end


---
function PalletAutoLoader.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDynamicMountTimeToMount", PalletAutoLoader.getDynamicMountTimeToMount)
end


---
function PalletAutoLoader.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", PalletAutoLoader)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", PalletAutoLoader)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", PalletAutoLoader)
end

---
function PalletAutoLoader:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    if self.isClient then
        local spec = self.spec_palletAutoLoader
        self:clearActionEventsTable(spec.actionEvents)

        if isActiveForInputIgnoreSelection then
            local state, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.ACTIVATE_OBJECT, self, PalletAutoLoader.actionEventToggleLoading, false, true, false, true, nil, nil, true, true)
            if not state then
                local _
                _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.IMPLEMENT_EXTRA4, self, PalletAutoLoader.actionEventToggleLoading, false, true, false, true, nil, nil, true, true)
            end
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL)
			spec.actionEventId = actionEventId;
            PalletAutoLoader.updateActionText(self);
        end
    end
end

---
function PalletAutoLoader.updateActionText(self)
		local spec = self.spec_palletAutoLoader
		
		if spec.objectsToLoadCount == 0 then
			text = g_i18n:getText("palletAutoLoader_nothingToLoad")
		else
			text = g_i18n:getText("palletAutoLoader_loadPallets") .. ": " .. spec.objectsToLoadCount
		end
		
		g_inputBinding:setActionEventText(spec.actionEventId, text)
		
		-- print("objectsToLoad");
		-- DebugUtil.printTableRecursively(spec.objectsToLoad,"_",0,2);
end

---
function PalletAutoLoader.actionEventToggleLoading(self, actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_palletAutoLoader
	
	-- beim einschalten müssen alle bereits vorhandenen objecte im trigger geprüft werden
	
	-- local copyList = {};
		-- for _, objectId in pairs(spec.objectsToLoad) do
		-- table.insert(copyList, objectId);
	-- end
	
	-- for _, objectId in pairs(copyList) do
		-- print("loadObject: " .. objectId);
		-- self:loadObject(objectId);
	-- end
	
	-- local _, objectId = pairs(spec.objectsToLoad);
	-- print( "objectId" );
	-- DebugUtil.printTableRecursively(objectId,"_",0,2)
	-- print( objectId );
	
	
	for _, object in pairs(spec.objectsToLoad) do
		self:loadObject(object);
		break;
	end
	
	PalletAutoLoader.updateActionText(self);
end

---
function PalletAutoLoader:onLoad(savegame)
    if self.isServer then
		--DebugUtil.printTableRecursively(self,"_",0,2)
		self.spec_palletAutoLoader = {}
        local spec = self.spec_palletAutoLoader
		spec.actionEvents = {}
		spec.objectsToLoad = {};
		spec.objectsToLoadCount = 0;

        spec.triggerId = self.xmlFile:getValue("vehicle.palletAutoLoader.trigger#node", nil, self.components, self.i3dMappings)
        if spec.triggerId ~= nil then
            addTrigger(spec.triggerId, "autoLoaderTriggerCallback", self);
        end
		
        spec.pickupTriggers = {}
		
        local i = 0
        while true do
            local pickupTriggerKey = string.format("vehicle.palletAutoLoader.pickupTriggers.pickupTrigger(%d)", i)
            if not self.xmlFile:hasProperty(pickupTriggerKey) then
                break
            end

            local entry = {}
            entry.node = self.xmlFile:getValue(pickupTriggerKey .. "#node", nil, self.components, self.i3dMappings)

            if entry.node ~= nil then
                table.insert(spec.pickupTriggers, entry)
				addTrigger(entry.node, "autoLoaderPickupTriggerCallback", self)
            end

            i = i + 1
        end

        spec.triggeredObjects = {}
        spec.numTriggeredObjects = 0
        spec.loadPlaces = {}

        local i = 0
        while true do
            local placeKey = string.format("vehicle.palletAutoLoader.loadPlaces.loadPlace(%d)", i)
            if not self.xmlFile:hasProperty(placeKey) then
                break
            end

            local entry = {}
            entry.node = self.xmlFile:getValue(placeKey .. "#node", nil, self.components, self.i3dMappings)

            if entry.node ~= nil then
                table.insert(spec.loadPlaces, entry)
            end

            i = i + 1
        end

        spec.loadPlaceOffset = self.xmlFile:getValue("vehicle.palletAutoLoader.loadPlaces#checkOffset", "0 0 0", true)
        spec.loadPlaceSize = self.xmlFile:getValue("vehicle.palletAutoLoader.loadPlaces#checkSize", "1 1 1", true)

        spec.supportedObject = self.xmlFile:getValue("vehicle.palletAutoLoader#supportedObject")

        spec.fillUnitIndex = self.xmlFile:getValue("vehicle.palletAutoLoader#fillUnitIndex")
        spec.maxObjects = self.xmlFile:getValue("vehicle.palletAutoLoader#maxObjects") or #spec.loadPlaces
        spec.useBales = self.xmlFile:getValue("vehicle.palletAutoLoader#useBales", false)
        spec.useTensionBelts = self.xmlFile:getValue("vehicle.palletAutoLoader#useTensionBelts", not GS_IS_MOBILE_VERSION)
		
		-- missing in xml
		spec.maxLoadHeight = 2;
    end
end

---
function PalletAutoLoader:onDelete()
    local spec = self.spec_palletAutoLoader

    if self.isServer then
        if spec.triggerId ~= nil then
            removeTrigger(spec.triggerId)
        end

        if spec.pickupTriggers ~= nil then
			for _, pickupTrigger in pairs(spec.pickupTriggers) do
				removeTrigger(pickupTrigger.node)
			end            
        end
    end
end

---
function PalletAutoLoader:getIsValidObject(object)
    local spec = self.spec_palletAutoLoader

	-- print("object.rootNode: " .. object.rootNode);
	-- DebugUtil.printTableRecursively(object.spec_mountable,"_",0,2)
						
	-- only when rootnode is object id it can be valid.
	if object.spec_mountable == nil or object.spec_mountable.componentNode ~= object.rootNode then
		return false;
	end
	
    if spec.supportedObject ~= nil then
        local objectFilename = object.configFileName or object.i3dFilename
        if objectFilename ~= nil then
            if not string.find(objectFilename, spec.supportedObject) then
                return false
            end
        else
            return false
        end
    end

    if object == self then
        return false
    end

    if spec.useBales and (not object:isa(Bale) or not object:getAllowPickup()) then
        return false
    end

    if not g_currentMission.accessHandler:canFarmAccess(self:getActiveFarm(), object) then
        return false
    end

    if spec.fillUnitIndex ~= nil and object.getFillType ~= nil and not self:getFillUnitSupportsFillType(spec.fillUnitIndex, object:getFillType()) then
        return false
    end

    return true
end

---
function PalletAutoLoader:getIsAutoLoadingAllowed()
    -- check if the vehicle has not fallen to side
    local _, y1, _ = getWorldTranslation(self.components[1].node)
    local _, y2, _ = localToWorld(self.components[1].node, 0, 1, 0)
    if y2 - y1 < 0.5 then
        return false
    end

    return true
end

---
function PalletAutoLoader:getDynamicMountTimeToMount(superFunc)
	return self:getIsAutoLoadingAllowed() and -1 or math.huge
end

---
function PalletAutoLoader:getFirstValidLoadPlace()
    local spec = self.spec_palletAutoLoader

	-- Hier die loading position und das loading objekt nehmen und die erste ladeposition dafür dynamisch suchen
	-- https://gdn.giants-software.com/documentation_scripting_fs19.php?version=engine&category=15&function=138
	-- overlapBox scheint zu prüfen, ob der angegebene Bereich frei ist
	
	local currentLoadHeigt = 0;
	while currentLoadHeigt  <= spec.maxLoadHeight do
	
		for i=1, #spec.loadPlaces do
			local loadPlace = spec.loadPlaces[i]
			local x, y, z = getWorldTranslation(loadPlace.node)
			local x, y, z = localToWorld(loadPlace.node, 0, currentLoadHeigt, 0);
			local rx, ry, rz = getWorldRotation(loadPlace.node)
			local offsetX, offsetY, offsetZ = unpack(spec.loadPlaceOffset)
			local sizeX, sizeY, sizeZ = unpack(spec.loadPlaceSize)
			
			-- collision mask : all bits except bit 13, 23, 30
			spec.foundObject = false
			--overlapBox(x + offsetX, y + offsetY + offsetZ, z, rx, ry, rz, sizeX / 2, sizeY / 2, sizeZ / 2, "autoLoaderOverlapCallback", self, 3212828671, true, false, true)
			overlapBox(x + offsetX, y + offsetY + (sizeY / 2), z + offsetZ, rx, ry, rz, sizeX / 2, sizeY / 2, sizeZ / 2, "autoLoaderOverlapCallback", self, 3212828671, true, false, true)

			if not spec.foundObject then
				return i, currentLoadHeigt
			end
		end
		
		currentLoadHeigt = currentLoadHeigt + 0.1
	end

    return -1, 0
end

---
function PalletAutoLoader:autoLoaderOverlapCallback(transformId)
    if transformId ~= 0 and getHasClassId(transformId, ClassIds.SHAPE) then
        local spec = self.spec_palletAutoLoader

        local object = g_currentMission:getNodeObject(transformId)
        if object ~= nil and object ~= self then
            spec.foundObject = true
        end
    end

    return true
end

---
function PalletAutoLoader:loadObject(object)

	-- print("self");
	-- DebugUtil.printTableRecursively(self,"_",0,3)
	
	--print(self.spec_palletAutoLoader.triggerId);
	--local myTrigger = g_currentMission:getNodeObject(self.spec_palletAutoLoader.triggerId);
	-- local myTrigger = getScale(self.spec_palletAutoLoader.triggerId);
	-- if myTrigger ~= nil then	
		-- print(self.spec_palletAutoLoader.triggerId);
		-- print(myTrigger);
		-- DebugUtil.printTableRecursively(myTrigger,"_",0,1)
	-- else
		-- print("trigger is null");
	-- end
	
	-- local x,y,z = getScale(self.spec_palletAutoLoader.triggerId);
	-- print("Scale x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getTranslation(self.spec_palletAutoLoader.triggerId);
	-- print("Translation x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getWorldRotation(self.spec_palletAutoLoader.triggerId);
	-- print("WorldRotation x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getWorldTranslation(self.spec_palletAutoLoader.triggerId);
	-- print("WorldTranslation x:"..x.." y:"..y.." z:"..z);
		
	-- local x,y,z = localToWorld(self.spec_palletAutoLoader.triggerId, -1, -1, -1);
	-- print("localToWorld x:"..x.." y:"..y.." z:"..z);
	
	-- places auslesen geht
	-- local x,y,z = getTranslation(self.spec_palletAutoLoader.loadPlaces[1].node);
	-- print("place 1 x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getTranslation(self.spec_palletAutoLoader.loadPlaces[2].node);
	-- print("place 2 x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getWorldTranslation(self.spec_palletAutoLoader.loadPlaces[1].node);
	-- print("world place 1 x:"..x.." y:"..y.." z:"..z);
	
	-- local x,y,z = getWorldTranslation(self.spec_palletAutoLoader.loadPlaces[2].node);
	-- print("world place 2 x:"..x.." y:"..y.." z:"..z);
	
	--local object = g_currentMission:getNodeObject(objectId)
	if object ~= nil then
		if self:getIsAutoLoadingAllowed() and self:getIsValidObject(object) then
			local spec = self.spec_palletAutoLoader
			if spec.triggeredObjects[object] == nil then
				if spec.numTriggeredObjects < spec.maxObjects then
					local firstValidLoadPlace, currentLoadHeigt = self:getFirstValidLoadPlace()
					if firstValidLoadPlace ~= -1 then
						local loadPlace = spec.loadPlaces[firstValidLoadPlace]
						local x,y,z = localToWorld(loadPlace.node, 0, currentLoadHeigt, 0);
						local objectNodeId = object.nodeId or object.components[1].node

						removeFromPhysics(objectNodeId)

						setTranslation(objectNodeId, x, y, z)
						setWorldRotation(objectNodeId, getWorldRotation(loadPlace.node))

						addToPhysics(objectNodeId)

						local vx, vy, vz = getLinearVelocity(self:getParentComponent(loadPlace.node))
						setLinearVelocity(objectNodeId, vx, vy, vz)

						spec.triggeredObjects[object] = 0
						spec.numTriggeredObjects = spec.numTriggeredObjects + 1

					-- print("Loaded: " .. object.rootNode);
					-- DebugUtil.printTableRecursively(object,"_",0,2)
						

						if spec.useTensionBelts and self.setAllTensionBeltsActive ~= nil then
							self:setAllTensionBeltsActive(false, false)
							--- PAUSE !!!
							
							self:setAllTensionBeltsActive(true, false)
						end
						
						if spec.objectsToLoad[object.rootNode] ~= nil then
							--table.remove(spec.objectsToLoad, objectId)
							spec.objectsToLoad[object.rootNode] = nil;
							spec.objectsToLoadCount = spec.objectsToLoadCount - 1;
						end
					end
				end
			end
		end
	end
	
end

---
function PalletAutoLoader:autoLoaderPickupTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
	-- print("autoLoaderPickupTriggerCallback: " .. otherActorId);
	if otherActorId ~= 0 then
		local object = g_currentMission:getNodeObject(otherActorId)
		if object ~= nil then
			if self:getIsAutoLoadingAllowed() and self:getIsValidObject(object) then
				local spec = self.spec_palletAutoLoader
				if onEnter then
					-- print("onEnter: " .. otherActorId);
					-- DebugUtil.printTableRecursively(spec.objectsToLoad,"_",0,2)
					if spec.objectsToLoad[object.rootNode] == nil then
						spec.objectsToLoad[object.rootNode] = object;
						spec.objectsToLoadCount = spec.objectsToLoadCount + 1;
						-- print("added object: " .. object.rootNode);
						-- DebugUtil.printTableRecursively(object,"_",0,2)
						PalletAutoLoader.updateActionText(self);
					end
					-- print("afterEnter: " .. object.rootNode);
					-- DebugUtil.printTableRecursively(spec.objectsToLoad,"_",0,2)
				elseif onLeave then
					-- print("onLeave: " .. object.rootNode);
					-- DebugUtil.printTableRecursively(spec.objectsToLoad,"_",0,2)
					if spec.objectsToLoad[object.rootNode] ~= nil then
						-- spec.objectsToLoad.otherActorId = nil;
						-- table.remove(spec.objectsToLoad, otherActorId)
						spec.objectsToLoad[object.rootNode] = nil;
						spec.objectsToLoadCount = spec.objectsToLoadCount - 1;
						PalletAutoLoader.updateActionText(self);
					end
					-- print("afterLeave: " .. otherActorId);
					-- DebugUtil.printTableRecursively(spec.objectsToLoad,"_",0,2)
				end
			end
		end
    end
end


--DebugUtil.printTableRecursively(object,"_",0,2)
--print(string.format("Geometry ID: %d", getGeometry(triggerId)));
-- local targetArea = g_currentMission:getNodeObject(triggerId)
-- if targetArea == nil then
-- targetArea = g_currentMission.nodeToObject[triggerId]
-- end
-- DebugUtil.printTableRecursively(targetArea,"_",0,2)

---
function PalletAutoLoader:autoLoaderTriggerCallback(triggerId, otherActorId, onEnter, onLeave, onStay, otherShapeId)
    local spec = self.spec_palletAutoLoader

    if onEnter then
        local object = g_currentMission:getNodeObject(otherActorId)
        if object ~= nil then
            if self:getIsValidObject(object) then
                if spec.triggeredObjects[object] == nil then
                    spec.triggeredObjects[object] = 0
                    spec.numTriggeredObjects = spec.numTriggeredObjects + 1
                end

                if spec.triggeredObjects[object] == 0 then
                    if object.addDeleteListener ~= nil then
                        object:addDeleteListener(self, "onDeletePalletAutoLoaderObject")
                    end
                end

                spec.triggeredObjects[object] = spec.triggeredObjects[object] + 1
            end
        end
    elseif onLeave then
        local object = g_currentMission:getNodeObject(otherActorId)
        if object ~= nil then
            if self:getIsValidObject(object) then
                if spec.triggeredObjects[object] ~= nil then
                    spec.triggeredObjects[object] = spec.triggeredObjects[object] - 1

                    if spec.triggeredObjects[object] == 0 then
                        spec.triggeredObjects[object] = nil
                        spec.numTriggeredObjects = spec.numTriggeredObjects - 1

                        if object.removeDeleteListener ~= nil then
                            object:removeDeleteListener(self)
                        end
                    end

                    if next(spec.triggeredObjects) == nil then
                        spec.currentPlace = 1
                    end
                end
            end
        end
    end
end

---
function PalletAutoLoader:onDeletePalletAutoLoaderObject(object)
    local spec = self.spec_palletAutoLoader

    if spec.triggeredObjects[object] ~= nil then
        spec.triggeredObjects[object] = nil
        spec.numTriggeredObjects = spec.numTriggeredObjects - 1

        if next(spec.triggeredObjects) == nil then
            spec.currentPlace = 1
        end
    end
end
